#include <stdio.h>
#include <stdlib.h>

#include "vrt.h"

#include "vcc_if.h"
#include "parse_encoding.h"
#include "wb.h"
//#include "cache/cache.h"

int
init_function(struct vmod_priv *priv, const struct VCL_conf *conf)
{
	return (0);
}

static const struct vmod_priv const null_blob[1] =
{
	{
		.priv = "",
		.len = 0,
		.free = NULL
	}
};

struct b64_state_s {
	uint32_t	magic;
#define B64_STATE_MAGIC	0xc0b64b64
	unsigned char	bytes[2];
	ssize_t	l;
};

struct 
	

static bool
b64_e(const enum encodings enc, const char *in,
size_t inlen, char *out, size_t outlen)
{

/*
 * Base64-encode *in (size: inlen) into *out, max outlen bytes. If there is
 * insufficient space, it will bail out and return -1. Otherwise, it will
 * null-terminate and return the used space.
 * The alphabet `a` defines... the alphabet. Padding is optional.
 * Inspired heavily by gnulib/Simon Josefsson (as referenced in RFC4648)
 *
 * XXX: tmp[] and idx are used to ensure the reader (and author) retains
 * XXX: a limited amount of sanity. They are strictly speaking not
 * XXX: necessary, if you don't mind going crazy.
 *
 * FIXME: outlenorig is silly. Flip the logic.
 */
static size_t
base64_encode (const enum encodings enc, const char *in,
               size_t inlen, char *out, size_t outlen)
{
        struct b64_alphabet *alpha = &b64_alphabet[enc];
        size_t outlenorig = outlen;
        unsigned char tmp[3], idx;

        if (outlen<4)
                return -1;

        if (inlen == 0) {
                *out = '\0';
                return (1);
        }

        while (1) {
                assert(inlen);
                assert(outlen>3);

                tmp[0] = (unsigned char) in[0];
                tmp[1] = (unsigned char) in[1];
                tmp[2] = (unsigned char) in[2];

                *out++ = alpha->b64[(tmp[0] >> 2) & 0x3f];

                idx = (tmp[0] << 4);
                if (inlen>1)
                        idx += (tmp[1] >> 4);
                idx &= 0x3f;
                *out++ = alpha->b64[idx];
                        
                if (inlen>1) {
                        idx = (tmp[1] << 2);
                        if (inlen>2)
                                idx += tmp[2] >> 6;
                        idx &= 0x3f;

                        *out++ = alpha->b64[idx];
                } else {
                        if (alpha->padding)
                                *out++ = alpha->padding;
                }

                if (inlen>2) {
                        *out++ = alpha->b64[tmp[2] & 0x3f];
                } else {
                        if (alpha->padding)
                                *out++ = alpha->padding;
                }

                /*
                 * XXX: Only consume 4 bytes, but since we need a fifth for
                 * XXX: NULL later on, we might as well test here.
                 */
                if (outlen<5)
                        return -1;

                outlen -= 4;

                if (inlen<4)
                        break;
                
                inlen -= 3;
                in += 3;
        }

        assert(outlen);
        outlen--;
        *out = '\0';
        return outlenorig-outlen;
}



static inline char
hexnibblelc(char c)
{
	return ((c > 9) ? ('a' + (c - 0xa)) : ('0' + c));
}


/*
 * en/decoders need to work for the string_list vargs case (implicit length) as
 * well as for the blob case (explicit length). Fortunately, the blob case is
 * not varadic, so we can assume just one length argument _or_ varags
 *
 * rblob: alloc additional whitespace so any result can also be (safely)
 *	  used as a string.
 */

static VCL_BLOB
hex_encode_va(struct ws *ws, struct vmod_priv *rblob, const ssize_t len,
    const char *in, va_list va)
{
	struct wb_s wb[1];
	char *p;
	ssize_t insz;
	int i;

	AN(rblob);

	if (in == vrt_magic_string_end)
		return null_blob;

	p = wb_create(ws, wb);
	if (p == NULL)
		return NULL;
	const char *e = wb_end(wb);

	if (len != -1)
		insz = len;
	else if (in == NULL)
		insz = 0;
	else
		insz = strlen(in);

	do {
		if (insz == 0)
			goto next;
		for (i = 0; i < insz && p + 1 < e; i++) {
			*p++ = hexnibblelc((in[i] & 0xf0) >> 4);
			*p++ = hexnibblelc( in[i] & 0x0f);
		}
		if (i != insz)
			goto err;

	  next:
		in = va_arg(va, const char *);
		if (in == vrt_magic_string_end)
			break;
		insz = (in != NULL) ? strlen(in) : 0;
	} while(1);

	wb_advanceP(wb, p);

	(void) wb_finish_blob(wb, rblob);

	if (rblob->len == 0)
		return null_blob;

	return rblob;

  err:
	wb_reset(wb);
	return NULL;
}

static VCL_BLOB
hex_encode(struct ws *ws, struct vmod_priv *rblob, const ssize_t len,
    const char *in, ...)
{
	va_list ap;
	VCL_BLOB r;
	va_start(ap, in);
	r = hex_encode_va(ws, rblob, len, in, ap);
	va_end(ap);
	return r;
}


/* Functions */

VCL_BLOB
vmod_decode(VRT_CTX, VCL_ENUM decs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(dec != _INVALID);

	return null_blob;
}
VCL_STRING
vmod_encode(VRT_CTX, VCL_ENUM encs, VCL_BLOB b) {
	enum encoding enc = parse_encoding(encs);
	struct vmod_priv tmp[1];
	const struct vmod_priv *r;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(enc != _INVALID);

	assert(enc == HEXLC);	// XXX
	r = hex_encode(ctx->ws, tmp, b->len, b->priv, vrt_magic_string_end);
	if (r == NULL)
		return NULL;
	return r->priv;
}

VCL_STRING __match_proto__(td_convert_transcode)
vmod_transcode(VRT_CTX, VCL_ENUM decs, VCL_ENUM encs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	enum encoding enc = parse_encoding(encs);
	struct vmod_priv tmp[1];

	va_list ap;
	VCL_BLOB r;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(dec != _INVALID);
	assert(enc != _INVALID);

	// XXX DO THE DECODE BIT

	assert(enc == HEXLC);	// XXX
	va_start(ap, p);
	r = hex_encode_va(ctx->ws, tmp, -1, p, ap);
	va_end(ap);
	if (r == NULL)
		return NULL;
	return r->priv;
}
VCL_BLOB
vmod_transcode_blob(VRT_CTX, VCL_ENUM decs, VCL_ENUM encs, VCL_BLOB b) {
	enum encoding dec = parse_encoding(decs);
	enum encoding enc = parse_encoding(encs);
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(dec != _INVALID);
	assert(enc != _INVALID);
	return b;
}
